/******************************************************************************
*  Copyright (c) 2002 Linux UPnP Internet Gateway Device Project              *
*  All rights reserved.                                                       *
*                                                                             *
*  This file is part of The Linux UPnP Internet Gateway Device (IGD).         *
*                                                                             *
*  The Linux UPnP IGD is free software; you can redistribute it and/or modify *
*  it under the terms of the GNU General Public License as published by       *
*  the Free Software Foundation; either version 2 of the License, or          *
*  (at your option) any later version.                                        *
*                                                                             *
*  The Linux UPnP IGD is distributed in the hope that it will be useful,      *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of             *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
*  GNU General Public License for more details.                               *
*                                                                             *
*  You should have received a copy of the GNU General Public License          *
*  along with Foobar; if not, write to the Free Software                      *
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
*                                                                             *
*                                                                             *
******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/sysinfo.h>
#include <string.h>
#include <time.h>
#include "tool.h"
#include "upnpd.h"
#include "action.h"
#include "nvram.h"
#include <net/if.h>
#include <stdarg.h>

//#define UPNP_DEBUG

extern struct GateEnv gate;
#define ONE_MILLION	(1000*1000)

int get_wan_phy_speed(unsigned int *speed);

#ifdef UPNP_DEBUG
void mBUG(char *format, ...);
#endif

int get_sockfd(void)
{
	 int sockfd = -1;
	
	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1){
			perror("user: socket creating failed");
			return (-1);
	}
	
	return sockfd;
}

char *getIPAddress(char *ifname)
{
	struct ifreq ifr;
	struct sockaddr_in *saddr;
	int fd;

	fd = get_sockfd();
	if (fd >= 0 )
	{
		memset(&ifr,0,sizeof(ifr));
		strcpy(ifr.ifr_name, ifname);
		ifr.ifr_addr.sa_family = AF_INET;
		if (ioctl(fd, SIOCGIFADDR, &ifr) == 0)
		{
			saddr = (struct sockaddr_in *)&ifr.ifr_addr;
			close(fd);
			return inet_ntoa(saddr->sin_addr);
		}
		else
		{
			//perror("get IP address");
			close(fd);
			return NULL;
		}
	}
	return NULL;
}
unsigned int getIPAddress2(char *ifname)
{
	struct ifreq ifr;
	struct sockaddr_in *saddr;
	int fd;

	fd = get_sockfd();
	if (fd >= 0 )
	{
		memset(&ifr,0,sizeof(ifr));
		strcpy(ifr.ifr_name, ifname);
		ifr.ifr_addr.sa_family = AF_INET;
		if (ioctl(fd, SIOCGIFADDR, &ifr) == 0)
		{
			saddr = (struct sockaddr_in *)&ifr.ifr_addr;
			close(fd);
			return saddr->sin_addr.s_addr;
		}
		else
		{
			//perror("get IP address");
			close(fd);
			return 0;
		}
	}
	return 0;
}

int getConnectionStatus(void)
{
    	if (access("/tmp/wan_uptime",F_OK)==0)
		return 1;  // connected
	else
		return 0;  //disconnected

}

/*
     2005.5.17 john
     direct get info from bcm adsl lib, not from console
*/
/** Typedefs. **/

typedef unsigned char   UINT8;
typedef unsigned short  UINT16;
typedef unsigned long   UINT32;


/* Check the interface status to see if the interface is up or down */
#if 0
int getConnectionStatus(void)
{

	FILE *fp;
	char tmp[1024];
	system("/usr/sbin/adslctl info --stats > /tmp/adsl-stats");
	fp=fopen("/tmp/adsl-stats","r");

	// default value is disconnected
	strcpy(gate.ConnectionStatus,"Disconnected");


	if(fp==NULL)
		return -1;

	fread(tmp,1024,1,fp);
	if(strstr(tmp,"Training Status:")==NULL){
		fclose(fp);
		return 0;
	}
	if(strstr(tmp,"Showtime"))
		strcpy(gate.ConnectionStatus,"Connected");
/*	else
		strcpy(gate.ConnectionStatus,"Disconnected");
*/

	return 1;
}
#endif


int getUptime(void)
{
	struct sysinfo info;
	sysinfo(&info);
	return (info.uptime - gate.startup_time);
}
#if 0
char * getDefaultConnectionService(void)
{
   char *t = malloc(strlen("NULL")+1);
   strcpy(t,"NULL");
   return t;
}

char * getConnectionType(void)
{
   char *t = malloc(strlen("IP_Routed")+1);
   strcpy(t,"IP_Routed");
   return t;
}
char * getPossibleConnectionTypes(void)
{
   return NULL;
}

char * getWANAccessType(void)
{
   return NULL;
}

#endif

int getLayer1MaxBitRate(void)
{
	unsigned long wan_speed = 0;
	unsigned int speed = 0;
	
	if (getConnectionStatus()==1) {
		get_wan_phy_speed(&speed);
		
		wan_speed = (unsigned long)(speed * ONE_MILLION);
#ifdef UPNP_DEBUG		
		mBUG("<%d>wan speed = <%lu>\n", __LINE__, wan_speed);
#endif
		gate.UpstreamSpeed   = wan_speed;
		gate.DownstreamSpeed = wan_speed;
	}
	else{
		gate.UpstreamSpeed   = 0;
		gate.DownstreamSpeed = 0;
	}
	
	return 1;
}

#if 0
int getLayer1MaxBitRate(void)
{
	FILE *fp;
	char tmp[1024];
	char *pt;

	system("/usr/sbin/adslctl info --stats > /tmp/adsl-stats");
	fp=fopen("/tmp/adsl-stats","r");
	if(fp==NULL)
		return -1;

	fread(tmp,1024,1,fp);
	/* get detial status */
	if((pt=strstr(tmp,"Rate (Kbps):"))==NULL){
		fclose(fp);
		return -1;
	}

	sscanf(pt,"%*[^:]: %d %d",&gate.DownstreamSpeed
			,&gate.UpstreamSpeed);

	fclose(fp);
	return 1;
}
#endif

#if 0
char * getPhysicalLinkStatus(void)
{
   char *t;

   t = malloc(strlen("Up")+1);
   strcpy(t,"Up");

   return t;
}
char * getWANAccessProvider(void)
{
   return NULL;
}
char * getMaximumActiveConnections(void)
{
   char *t = malloc(2);
   t[0] = '0';
   return t;
}
char * getNumberOfActiveConnections(void)
{
   // don't know how to get this
   char *t = malloc(2);
   t[0] = '0';
   return t;
}

char * getActiveConnectionDeviceContainer(void)
{
   return NULL;
}

char * getActiveConnectionServiceID(void)
{
   return NULL;
}

#endif
/* it seems that getTotalBytesSent(void) & getTotalBytesReceived(void)
   are not being used */

unsigned long getTotalBytesSent(void)
{
	char dev[15];
	char line[512];
	FILE *stream;
        unsigned long bytes=0, total=0;

	/* Read sent from /proc */
	stream = fopen ( "/proc/net/dev", "r" );
	
	if ( stream != NULL )
	{
		while ( fgetc ( stream ) != '\n' );
		while ( fgetc ( stream ) != '\n' );

		while ( fgets(line, sizeof(line),stream) )
		{
			sscanf ( line, "%[^:]:%*u %*u %*u %*u %*u %*u %*u %*u %lu %*u %*u %*u %*u %*u %*u %*u\n", dev, &bytes );
                        if ( strcmp ( dev, gate.ext_ifname)==0 )
                                total += bytes;
		}
		fclose ( stream );
	}
	
   return total;


}

unsigned long getTotalBytesReceived(void)
{
	char dev[15];
	char line[512];	
	FILE *stream;
        unsigned long bytes=0, total=0;

	/* Read sent from /proc */
	stream = fopen ( "/proc/net/dev", "r" );
	if ( stream != NULL )
	{
		while ( fgetc ( stream ) != '\n' );
		while ( fgetc ( stream ) != '\n' );

		while ( fgets(line, sizeof(line),stream) )
		{
                        sscanf ( line, "%[^:]:%lu %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u\n", dev, &bytes );
                        if ( strcmp ( dev, gate.ext_ifname)==0 )
                                total += bytes;
		}
		fclose ( stream );
	}
   return total;

}

int getNATEnabled(void)
{
#ifdef NVRAM
	char *fw_nat;

	fw_nat=nvram_get("fw_nat");
	if(fw_nat && *fw_nat=='0')
	{
	    free(fw_nat);
		return 0;
	}
	if(fw_nat && *fw_nat=='1')
	{
	    free(fw_nat);
		return 1;
	}
	else
	{
	    if(fw_nat)
	        free(fw_nat);
	        
		return -1;
	}
#else

	return 1;
#endif

}



unsigned long getTotalPacketsSent(void)
{

	char dev[15];
	char line[512];
	FILE *stream;
        unsigned long pkt=0, total=0;

	/* Read sent from /proc */
	stream = fopen ( "/proc/net/dev", "r" );
	if ( stream != NULL )
	{
		while ( fgetc ( stream ) != '\n' );
		while ( fgetc ( stream ) != '\n' );

		while (fgets(line, sizeof(line),stream))
		{
                        sscanf ( line, "%[^:]:%*u %*u %*u %*u %*u %*u %*u %*u %*u %lu %*u %*u %*u %*u %*u %*u\n", dev, &pkt );
                        if ( strcmp ( dev, gate.ext_ifname)==0 )
                                total += pkt;
		}
		fclose ( stream );
	}
   return total;
}

unsigned long getTotalPacketsReceived(void)
{

	char dev[15];
	char line[512];
        FILE *stream;
        unsigned long pkt=0, total=0;

        /* Read sent from /proc */
      stream = fopen ( "/proc/net/dev", "r" );
        if ( stream != NULL )
        {
                while ( fgetc ( stream ) != '\n' );
                while ( fgetc ( stream ) != '\n' );

                while ( fgets(line, sizeof(line),stream) )
                {
                        sscanf ( line, "%[^:]:%*u %lu %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u\n", dev, &pkt );
										
                        if ( strcmp ( dev, gate.ext_ifname )==0 )
                                total += pkt;

                }
             
                fclose ( stream );
        }

	//printf(" received  %lu\n",total);
	return total;

}

int getEnabledForInternet(void)
{
      return 1;
}

int getDHCPServerConfigurable(void)
{
   return 1; // yes
}

int getDHCPRelay(void)
{
   return 1; // yes
}

char * getSubnetMask(void)
{
    char *netmask="255.255.255.0";
    return netmask;
}

char * getIPRouters(void)
{
   return NULL;
}

char * getDNSServers(void)
{
      return NULL;
}

char * getDomainName(void)
{
   return NULL;
}

char * getMinAddress(void)
{
   return NULL;
}

char * getMaxAddress(void)
{
   return NULL;
}

char * getReservedAddress(void)
{
   return NULL;
}

char * getHostName(void)
{
	char *host="UPNP_HOST";
	return  host;
}

int substr(char *infile, char *outfile, char *str_from, char *str_to)
{
	#define TMP_FILE "/tmp/upnp_tmp"
	FILE *fpi, *ftmp;
        char pathi[256], patho[256];
        char buffi[4096], buffo[4096];
	 char bufcmd[256];
        int len_buff, len_from, len_to;
        int i, j;

        sprintf(pathi, "%s", infile);
        if ((fpi = fopen(pathi,"r")) == NULL)
        {
//                printf("input file can not open\n");
                return (-1);
        }

        sprintf(patho, "%s",TMP_FILE);
        if ((ftmp = fopen(patho,"w")) == NULL) {
//                printf("output file can not open\n");
                fclose(fpi);
                return (-1);
        }

        len_from = strlen(str_from);
        len_to   = strlen(str_to);

        while (fgets(buffi, 4096, fpi) != NULL)
        {
                len_buff = strlen(buffi);
                for (i=0, j=0; i <= len_buff-len_from; i++, j++)
                {
                        if (strncmp(buffi+i, str_from, len_from)==0)
                        {
                                strcpy (buffo+j, str_to);
                                i += len_from - 1;
                                j += len_to - 1;
                        } else
                        *(buffo + j) = *(buffi + i);
                }
                strcpy(buffo + j, buffi + i);
                fputs(buffo, ftmp);
        }

        fclose(ftmp);
        fclose(fpi);
	 sprintf(bufcmd,"cp -f %s %s",TMP_FILE,outfile);
	 system(bufcmd);
	 unlink(TMP_FILE);
        return (0);
}
/*******************************************************************************
* NAME: get_wan_phy_speed
* INPUTS: 
* OUTPUTS:
* RETURNS:
*       OK   - on success 0
*       FAIL - on error 1
*
*******************************************************************************/
int get_wan_phy_speed(unsigned int *speed)
 {
 	char tmpBuf[256]={0};
	FILE *fp;
	int len=0;
	char *p;
	int status = 0;
	char *adsl_speed=nvram_get("adsl_speed");
	//raul add temp speed
			if(adsl_speed)
				{					
				*speed=atoi(adsl_speed)/1000;
				}
			else	
				*speed = 10;
			return 0;	
	fp = fopen("/proc/eth_status","r");
	if(!fp){
		printf("fopen failed\n");
		return 1;
	}
	len = fread(tmpBuf, 1, sizeof(tmpBuf)-1, fp);
	fclose(fp);
	if(len <= 0)
		return 1;
	/* in our board, wan is 4 port*/	
    p = strstr(tmpBuf, "=");
    if(!p)
    	return 1;
    
    status = *(p+1) - '0';

#ifdef UPNP_DEBUG
	mBUG("<%d>wan speed=<%d>\n", __LINE__, status);    
#endif
	
	switch(status){
		case 0:
			*speed = 10;
			break;
		case 1:
			*speed = 100;
			break;
		case 2:
			*speed = 1000;
		default:
			break;
		}

	return 0;
 }

void mBUG(char *format, ...)
{
    va_list args;
    FILE *fp;

    fp = fopen("/var/upnp_test", "a+");
    if (!fp){
        return ;
    }
    va_start(args, format);
    vfprintf(fp, format, args);
    va_end(args);
    fprintf(fp, "\n");
    fflush(fp);
    fclose(fp);
    system("/bin/chmod 777 /var/upnp_test");
}
