#ifndef _XT_TD_H_target
#define _XT_TD_H_target



/*
** Define some types
*/
typedef void VOID;
typedef char CHAR;
typedef unsigned char U_CHAR;
typedef int INT;
typedef unsigned int U_INT;
typedef long LONG;
typedef unsigned long U_LONG;
typedef unsigned int __U32;
typedef unsigned int __BE32;
typedef unsigned short U_SHORT;
typedef unsigned short __U16;
typedef unsigned short __BE16;

/*
** Some Macro
*/
#define EIGHTH_K 128
#define QTR_K 256
#define HALF_K 512

#ifndef IFNAMSIZ
	#define IF_NAME_LEN 16
#else
	#define IF_NAME_LEN IFNAMSIZ
#endif /* IFNAMSIZ */
#define IP_LEN 4
#define MAC_LEN 6

#define DOT_IP_LEN 16
#define COLON_MAC_LEN 32
#define CHAR_PORT_LEN 8
#define CHAR_PROTOCOL_LEN 8
#define COMMAND_LEN 256

#define NOISES_NUM 8
#define ELIMINATE_NOISES_SCALE 5      /* 1/2^4 */
#define SMALL_PACKET_LEN 384
#define MEDIUM_PACKET_LEN 768
#define BIG_PACKET_LEN 1024           /* NEED TEST */
#define LAN_PC_NUM 256
#define TRACK_TIME 100                /* 0.1seconds */
#define SLEEP_TIME 50                 /* 0.1seconds */
#define LONG_SLEEP_TIME 600           /* 0.1seconds */
#define TCP_TRACK_DELAY 20            /* 0.1second */
#define LONG_LIFE_TIME 600           /* 0.1seconds */
#define LARGE_TRAFFIC 30720           /* 240KBit, in TRACK_TIME */
#define MAYBE_GOOD_PORT_LIMIT 2048    /* we consider port less than 2048 maybe good */
#define HTTP_MAX_PORT 6000
#define PC_SAFE_NEW_SESSIONS 150           /* new sessions created in TRACK TIME */
#define FEWER_PC_SAFE_NEW_SESSIONS 60      /* new sessions created in TRACK TIME */
#define PC_EXPIATION_TIMES 18              /* 3minutes 18*TRACK_TIME */
#define PC_MAX_TRAFFIC_GAGE(x) ((x*19)>>5)
#define PC_MAX_SESSION_GAGE(x) ((x*13)>>4)

#define VOIP_MAX_INTERVAL 8           /* 0.1seconds */
#define VOIP_PKTS_LIMIT 10            /* 10 pkts in TRACK_TIME */

#define GAMING_MAX_INTERVAL 15        /* 0.1seconds */
#define GAMING_LEAST_PKTS 15          /* PKTS/TRACK_TIME */
#define GAMING_BIGPKTS_SCALE 3        /* 2^n multiple */
#define GAMING_MAX_TRAFFIC 327680     /* 2560KBit/TRACK_TIME */

#define P2P_MAX_BAD_SESSIONS 100      /*  */
#define P2P_PKTS_LIMIT 10             /* 10 pkts in TRACK_TIME */

#define MAXGET(x, y) (x>=y?x:y)
#define MINGET(x, y) (x<=y?x:y)
#define OTHERGET(x, y, z) (x!=y?x:z)

/*
** Define some session types
*/
#define MARK_BASE 0x800
#define TYPE2MARK(x) (x+MARK_BASE)
enum session_type {
	FIRST_PKT_SESSION,     /* the first packet come */
	EARLY_UNKNOWN_SESSION, /* not the first packet recorded but has not be determined */
	UNSAFE_SESSION,
	MAYBE_GOOD_SESSION,

	/* the first result detected */
	UNKNOWN_SESSION,       /* maturely unknown but short life session */
	BAD_UNKNOWN_SESSION,   /* maturely unknow but long life session */
	VOIP_SESSION,          /* SIP session */
	GAMING_SESSION,        /* GAMING session */
	P2P_SESSION,           /* P2P session */
	HTTP_SESSION,
	D_UNKNOWN_SESSION,
	D_BAD_UNKNOWN_SESSION,
	D_BAD_HTTP_SESSION,

	/* the final determinate result */
	D_VOIP_SESSION,
	D_GAMING_SESSION,
	D_P2P_SESSION,
	D_DNS_SESSION,
	D_MAIL_SESSION,
	D_FTP_SESSION,
	MAX_SESSION           /* the tail of session type */
};

enum queue_prio {
	HIGH_PRIO,
	NORMAL_PRIO,
	LOW1_PRIO,
	LOW2_PRIO,
    Q_MAX_PRIO,
};

/*user space application should flow this defination...*/
/*be sure SELF_DEFINE_MARK_BASE > MAX_SESSION !!!*/
#define SELF_DEFINE_MARK_BASE 0x880
#define PRIO_SELF_DEFINE_MARK_NUM 16
#define PRIO_SELF_DEFINE_MARK_BASE SELF_DEFINE_MARK_BASE
#define PRIO2MARK(prio) (prio+PRIO_SELF_DEFINE_MARK_BASE)
#define RATE_SELF_DEFINE_MARK_BASE SELF_DEFINE_MARK_BASE+PRIO_SELF_DEFINE_MARK_NUM
#define INDEX2MARK(index) (index+RATE_SELF_DEFINE_MARK_BASE)

#define USRDEF_HW_NAT_PASS  (SELF_DEFINE_MARK_BASE-MARK_BASE+HIGH_PRIO)
#define USRDEF_HW_NAT_UNPASS (SELF_DEFINE_MARK_BASE-MARK_BASE+Q_MAX_PRIO)

/* ALG mark must be bigger than  RATE_SELF_DEFINE_MARK_BASE+MAX_USER_RULES*/
#define SIP_ALG_MARK 0x900

/*
** Differentiate sending or replying packet
*/
enum original_reply{
	ORIGINAL_DIR,
	REPLY_DIR,
	MAX_DIR
};

/*
** Structure for tracking one session
*/
struct track_record {
	U_LONG initial_time;         /* the time when first data packet comes */
	U_INT start_time;            /* the time when start to track this time */
	U_INT time_stamp;            /* the time when current data packet comes */
	U_INT max_intvl[NOISES_NUM]; /* record NOISES_NUM maximal interval */
	U_INT small_pkts[MAX_DIR];   /* summation sending or replying packets whose length <SMALL_PACKET_LEN */
	U_INT medium_pkts[MAX_DIR];  /* summation sending or replying packets whose length >=SMALL_PACKET_LEN and length=<BIG_PACKET_LEN */
	U_INT big_pkts[MAX_DIR];     /* summation sending or replying packets whose length >BIG_PACKET_LEN */
	U_LONG traffic;              /* the total traffic in this TRACK_TIME */
	enum session_type type;      /* session type */
};

/*
** Differentiate sending or replying packet
*/
enum dev_dir {
	WAN2LAN,
	LAN2WAN,
	MAX_DEV_DIR
};

/*
** Statistic for types
*/
struct last_stat {
	U_LONG pkt_num[MAX_SESSION]; /* packets number */
	U_LONG traffic[MAX_SESSION]; /* (bytess) ummation of all packets length */
};

/*
** Statistic for per PC
*/
struct lan_pc_info {
	INT index;
	enum dev_dir d_dir;
};

/*
** Record a PCs info
*/
struct pc_record {
	U_LONG time_stamp;
	CHAR s_limit;
	U_INT term_s;  /* new sessions per TRACK_TIME */
	U_LONG term_t; /* total traffic per TRACK_TIME */
	U_INT alive_s[MAX_SESSION];
};

/*
** Record TD state
*/
struct startup_flag {
	U_LONG time_stamp;
	CHAR reset;
	CHAR fore_stat;
};

/*
** for HW_NAT
*/
struct hnat_param {
	U_CHAR dev_wan_mac[MAC_LEN];
	U_CHAR dev_lan_mac[MAC_LEN];
	U_CHAR external_mac[MAC_LEN];
	U_CHAR internal_mac[MAC_LEN];
	__BE32 dev_wan_ip;
	__BE32 external_ip;
	__BE32 internal_ip;
	__BE16 dev_wan_port;
	__BE16 external_port;
	__BE16 internal_port;
	__BE16 protocol;
	U_CHAR wan_vlan_act;
	U_SHORT wan_vlan_id;
	U_CHAR wan_pppoe_act;
	U_SHORT wan_pppoe_id;
	U_INT wan_dst_port;
	U_INT wan_act_dp;
	U_CHAR lan_vlan_act;
	U_SHORT lan_vlan_id;
	U_CHAR lan_pppoe_act;
	U_SHORT lan_pppoe_id;
	U_INT lan_dst_port;
	U_INT lan_act_dp;
};

/*
** for netlink
*/
enum {
	REQUEST_NL_TD,
	READY_NL_TD,
	RESET_NL_TD,
	MAX_NL_TD
};



/*
** Define some parameter types
*/
enum param_type {
	P_SESS_DOWN_UP,
	P_WAN_LAN_WL,
	P_SESSIONS,
	P_DOWNSTREAM,
	P_UPSTREAM,
	P_WAN_MODE,
	P_REMOTE_MAC,
	P_WAN_IF,
	P_WAN_MAC,
	P_WAN_IP,
	P_LAN_IF,
	P_LAN_MAC,
	P_LAN_IP,
	P_LAN_SUBNET,
	P_WL_IF,
	P_HW_NAT,
	P_TRACK_TIME,
	P_SLEEP_TIME,
	P_LONG_SLEEP_TIME,
	P_TCP_TRACK_DELAY,
	P_LONG_LIFE_TIME,
	P_LARGE_TRAFFIC,
	P_MAYBE_GOOD_PORT_LIMIT,
	P_ELIMINATE_NOISES_SCALE,
	MAX_PARAM
};

enum {
	NET_LAN,
	MASK_LAN,
	LAN_MAX
};

struct xt_td_target_info {
	U_LONG downstream;
	U_LONG upstream;
	U_INT sessions;
	CHAR wan_mode[EIGHTH_K];
	U_SHORT pppoe_id;
	U_CHAR remote_mac[MAC_LEN];
	CHAR wan_if[IF_NAME_LEN];
	U_CHAR wan_mac[MAC_LEN];
	__BE32 wan_ip;
	CHAR lan_if[IF_NAME_LEN];
	U_CHAR lan_mac[MAC_LEN];
	__BE32 lan_ip;
	__BE32 lan_subnet[LAN_MAX];
	CHAR wl_if[IF_NAME_LEN];
	INT hwnat_en;
	U_INT track_time;
	U_INT sleep_time;
	U_INT long_sleep_time;
	U_INT tcp_track_delay;
	U_INT long_life_time;
	U_LONG large_traffic;
	U_INT maybe_good;
	U_INT eliminate_noises;
};



#endif /* _XT_TD_H_target */
