/*
** Copyright ? 2005 SerComm Corporation.
** All Rights Reserved. SerComm Corporation reserves the right to make changes
** to this document without notice. SerComm Corporation makes no warranty,
** representation or guarantee regarding the suitability of its products for
** any particular purpose. SerComm Corporation assumes no liability arising out
** of the application or use of any product or circuit. SerComm Corporation
** specifically disclaims any and all liability, including without limitation
** consequential or incidental damages; neither does it convey any license under
** its patent rights, nor the rights of others.
*/

/*
** Traffic Detection
*/
#ifdef _SMART_QOS_



#include <linux/types.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/inetdevice.h>
#include <linux/if_vlan.h>
#include <linux/if_pppox.h>
#include <linux/if_arp.h>
#include <net/protocol.h>
#include <net/arp.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter/nf_conntrack_common.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_TD.h>
#include <linux/netlink.h>
#include <linux/spinlock.h>
#include <net/sock.h>
#include <net/route.h>
#include "../../nat/hw_nat/hwnat_ioctl.h"
#include "../../nat/hw_nat/ac_ioctl.h"



#if 0
	#define PRINT printk
	//#define X_DBG PRINT
	#define X_DBG(format, args...)

	#define DUMP_H_TUPLE(tp) \
		do{ \
			PRINT("%s,%d: %s <%u.%u.%u.%u:%u -> %u.%u.%u.%u:%u>\n", \
					__FUNCTION__, __LINE__, IPPROTO_TCP==tp->dst.protonum?"TCP":(IPPROTO_UDP==tp->dst.protonum?"UDP":"OTHER"), \
					((U_CHAR *)&(tp->src.ip))[3], ((U_CHAR *)&(tp->src.ip))[2], ((U_CHAR *)&(tp->src.ip))[1], ((U_CHAR *)&(tp->src.ip))[0], \
					IPPROTO_TCP==tp->dst.protonum?tp->src.u.tcp.port:(IPPROTO_UDP==tp->dst.protonum?tp->src.u.udp.port:0), \
					((U_CHAR *)&(tp->dst.ip))[3], ((U_CHAR *)&(tp->dst.ip))[2], ((U_CHAR *)&(tp->dst.ip))[1], ((U_CHAR *)&(tp->dst.ip))[0], \
					IPPROTO_TCP==tp->dst.protonum?tp->dst.u.tcp.port:(IPPROTO_UDP==tp->dst.protonum?tp->dst.u.udp.port:0)); \
	}while(0)

	#define DUMP_STAT(stat) \
		do{ \
			enum dev_dir d_dir; \
			enum session_type type; \
			for(d_dir=WAN2LAN; d_dir<MAX_DEV_DIR; d_dir++) \
			{ \
				PRINT("%s,%d: stat[%02d]={\n", __FUNCTION__, __LINE__, d_dir); \
				for(type=FIRST_PKT_SESSION; type<MAX_SESSION; type++) \
				{ \
					PRINT("[%02d]={%10lupkts, %10lubytes}\n", \
							type, \
							stat[d_dir].pkt_num[type], \
							stat[d_dir].traffic[type]); \
				} \
				PRINT("}device[%02d]\n", d_dir); \
			} \
			PRINT("\n"); \
		}while(0)

	#define DUMP_TRACK(trk_rcd) \
		do{ \
			INT i; \
			PRINT("%s,%d: trk_rcd={\n\t %u,\n\t %u,\n\t %u,\n", \
					__FUNCTION__, __LINE__, jiffies_to_tenthsecs(trk_rcd->initial_time), trk_rcd->start_time, trk_rcd->time_stamp); \
			PRINT("\t {"); \
			for(i=0; i<NOISES_NUM; i++) \
			{ \
				PRINT("%d, ", trk_rcd->max_intvl[i]); \
			} \
			PRINT("}, \n"); \
			PRINT("\t {%u, %u},\n", trk_rcd->small_pkts[ORIGINAL_DIR], trk_rcd->small_pkts[REPLY_DIR]); \
			PRINT("\t {%u, %u},\n", trk_rcd->medium_pkts[ORIGINAL_DIR], trk_rcd->medium_pkts[REPLY_DIR]); \
			PRINT("\t {%u, %u},\n", trk_rcd->big_pkts[ORIGINAL_DIR], trk_rcd->big_pkts[REPLY_DIR]); \
			PRINT("\t %lu, \n", trk_rcd->traffic); \
			PRINT("\t %d \n", trk_rcd->type); \
			PRINT("}\n"); \
		}while(0)

	#define DUMP_PC_RCD(p_pcrcd) \
		do{ \
			INT i; \
			PRINT("p_pcrcd={%3u, %3u, %lu\n\t{", p_pcrcd->s_limit, p_pcrcd->term_s, p_pcrcd->term_t); \
			for(i=0; i<MAX_SESSION; i++) \
			{ \
				PRINT("%3u,", p_pcrcd->alive_s[i]); \
			} \
			PRINT("}\n}\n"); \
		}while(0)

	#define DUMP_ALL_PC_RCD(pc_rcd) \
		do{ \
			INT i; \
			INT j; \
			PRINT("%s,%d: dump all pc's record\n", __FUNCTION__, __LINE__); \
			for(i=0; i<LAN_PC_NUM; i++) \
			{ \
				PRINT("pc_rcd[%03d]={%3u, {", i, pc_rcd[i].term_s); \
				for(j=0; j<MAX_SESSION; j++) \
				{ \
					PRINT("%3u,", pc_rcd[i].alive_s[j]); \
				} \
				PRINT("}}\n"); \
			} \
		}while(0)

#else
	#define PRINT(format, args...)
	#define X_DBG(format, args...)
	#define DUMP_H_TUPLE(tp)
	#define DUMP_STAT(stat)
	#define DUMP_TRACK(trk_rcd)
	#define DUMP_PC_RCD(p_pcrcd)
	#define DUMP_ALL_PC_RCD(pc_rcd)
#endif



MODULE_LICENSE("SERCOMM");
MODULE_AUTHOR("SERCOMM");
MODULE_DESCRIPTION("traffic detection target module");

#define EXISTENT_SESSIONS(p_pcrcd) \
	do{ \
		enum session_type type; \
		INT sessions=0; \
		for(type=FIRST_PKT_SESSION; type<MAX_SESSION; type++) \
		{ \
			sessions+=p_pcrcd[type]; \
		} \
		return sessions; \
	}while

struct startup_flag startup={0};
static struct xt_td_target_info td_info={0};
static struct pc_record pc_rcd[LAN_PC_NUM]={{0}};
static struct last_stat stat_tab[MAX_DEV_DIR]={{{0}}};
static struct sockaddr_pppox s_pppox;
static rwlock_t startup_lock; /* rw lock */
static rwlock_t stat_lock;
//static rwlock_t record_lock;
static rwlock_t pppox_lock;

/*
** Convert jiffies to tenthseconds.
** Avoid unnecessary multiplications/divisions in the two most common HZ cases:
*/
#define FULL_JIFFIES 4294967295UL /* 0xFFFFFFFF */
#define TENTHSEC_PER_SEC 10
#define TENTHSEC2SEC(x) (x/TENTHSEC_PER_SEC)
static U_INT jiffies_to_tenthsecs(const U_LONG j)
{
#if HZ <= TENTHSEC_PER_SEC && !(TENTHSEC_PER_SEC % HZ)
	return (TENTHSEC_PER_SEC / HZ) * j;
#elif HZ > TENTHSEC_PER_SEC && !(HZ % TENTHSEC_PER_SEC)
	return (j + (HZ / TENTHSEC_PER_SEC) - 1)/(HZ / TENTHSEC_PER_SEC);
#else
	return (j * TENTHSEC_PER_SEC) / HZ;
#endif
}

static U_INT relative_tenthsecs(U_LONG x, U_LONG y)
{
	if(y==0)
		return jiffies_to_tenthsecs(x);
	else if(time_after_eq(x, y))
	{
		if(x>=y)
			return jiffies_to_tenthsecs(x-y);
		else
			return jiffies_to_tenthsecs(x+FULL_JIFFIES-y);
	}
	else
		return 0;
}

static INT get_devip_by_name(const CHAR *name, __BE32 *ip)
{
	struct net_device *nd=NULL;
	if((nd=dev_get_by_name(name))!=NULL) 
	{
		*ip=inet_select_addr(nd, 0, 0);
		dev_put(nd);
	}
	else
		return -1;
	return 0;
}

static INT get_devmac_by_name(const CHAR *name, U_CHAR *mac)
{
	struct net_device *nd=NULL;
	if((nd=dev_get_by_name(name))!=NULL) 
	{
		memcpy(mac, nd->dev_addr, MAC_LEN);
		dev_put(nd);
	}
	else
		return -1;
	return 0;
}

static INT get_mac_in_arp(struct net_device *dev, __BE32 ip, U_CHAR *mac)
{
	struct neighbour *n=__neigh_lookup(&arp_tbl, &ip, dev, 0);
	
	X_DBG("%s,%d: lookup arp_tbl for <%s, %u.%u.%u.%u>\n", __FUNCTION__, __LINE__, dev->name, NIPQUAD(ip));
	if(!n)
	{
		X_DBG("%s,%d: can not find arp entry for device<%s>\n", __FUNCTION__, __LINE__, dev->name);
		return -1;
	}

	memcpy(mac, n->ha, MAC_LEN);
	
	X_DBG("%s,%d: ip<%u.%u.%u.%u> mac<%02X:%02X:%02X:%02X:%02X:%02X>\n",
			__FUNCTION__, __LINE__,
			NIPQUAD(*(__BE32*)n->primary_key),
			mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
	return 0;
}

static INT extract_pc_info(struct ip_conntrack_tuple *tuple,
		                       struct xt_td_target_info *info,
		                       struct lan_pc_info *pc_info)
{
	if(info->lan_subnet[NET_LAN]==0)
		return -1;

	if((tuple->src.ip & info->lan_subnet[MASK_LAN]) == info->lan_subnet[NET_LAN])
	{
		pc_info->index=tuple->src.ip & ~(info->lan_subnet[MASK_LAN]);
		pc_info->d_dir=LAN2WAN;
	}
	else if((tuple->dst.ip & info->lan_subnet[MASK_LAN]) == info->lan_subnet[NET_LAN])
	{
		pc_info->index=tuple->dst.ip & ~(info->lan_subnet[MASK_LAN]);
		pc_info->d_dir=WAN2LAN;
	}
	else
		return -1;

	X_DBG("conntrack protocol<%d> dir<%d> <%lx -> %lx> lan_subnet<%lx/%lx> pc_info={%d, %d}\n",
			tuple->dst.protonum, tuple->dst.dir, tuple->src.ip, tuple->dst.ip,
			info->lan_subnet[NET_LAN], info->lan_subnet[MASK_LAN], pc_info->index, pc_info->d_dir);

	if(pc_info->index<=0 || pc_info->index>=LAN_PC_NUM)
		return -1;
	else
		return 0;
}

static INT update_pc_alive_s(struct pc_record *p_pcrcd,
		                      struct track_record *trk_rcd,
		                      enum session_type cur_type,
		                      INT flag)
{
	if(flag==-1 && p_pcrcd->alive_s[trk_rcd->type]>0)
		p_pcrcd->alive_s[trk_rcd->type]--;
	else if(flag==0)
	{
		if(cur_type>=MAX_SESSION || cur_type<0 || trk_rcd->type==cur_type)
			return -1;
		if(p_pcrcd->alive_s[trk_rcd->type]>0)
			p_pcrcd->alive_s[trk_rcd->type]--;
		trk_rcd->type=cur_type;
		p_pcrcd->alive_s[trk_rcd->type]++;
	}
	else if(flag==1)
		p_pcrcd->alive_s[trk_rcd->type]++;
	else
		return -1;
	
	return 0;
}

/*
==========<Flow Table Entry=950>===============
Information Block 1: 20CE1296
SIP: AC150430
DIP: AC150402
SP|DP: 0014085B
Information Block 2: 00010003
NewSIP: AC150430
NewDIP: C0A80003
NewSP|NewDP: 0014085B
VL1|DMAC_Hi: 0001001B
DMAC_Lo: 210B88B3
PPPoE|SMAC_Hi: 000000C0
SMAC_Lo: 02112234
Resv|SNAP: 00000000
Resv|VL2: 00000000
Resv: 00000000
Resv: 00000000
-----------------<Flow Info>------------------
ORG: 172.21.4.48:20->172.21.4.2:2139
NEW: 172.21.4.48:20->192.168.0.3:2139
DMAC=00:1B:21:0B:88:B3 SMAC=00:C0:02:11:22:34
=========================================

==========<Flow Table Entry=1418>===============
Information Block 1: 20CE1296
SIP: C0A80003
DIP: AC150430
SP|DP: 085B0014
Information Block 2: 00010005
NewSIP: AC150402
NewDIP: AC150430
NewSP|NewDP: 085B0014
VL1|DMAC_Hi: 0002001C
DMAC_Lo: C4991E7A
PPPoE|SMAC_Hi: 000000C0
SMAC_Lo: 02112235
Resv|SNAP: 00000000
Resv|VL2: 00000000
Resv: 00000000
Resv: 00000000
-----------------<Flow Info>------------------
ORG: 192.168.0.3:2139->172.21.4.48:20
NEW: 172.21.4.2:2139->172.21.4.48:20
DMAC=00:1C:C4:99:1E:7A SMAC=00:C0:02:11:22:35
=========================================
*/
static INT (*is_wl_client)(struct net_device *dev, U_CHAR *mac)=NULL;
static INT (*add_hwnat)(struct hwnat_tuple *opt)=NULL;
static INT (*del_hwnat)(struct hwnat_tuple *opt)=NULL;
static INT (*get_hwnat_entrynum)(INT state)=NULL;

VOID init_wl_funp(INT (*wl_fun)(struct net_device *dev, U_CHAR *mac))
{
	is_wl_client=wl_fun;
}
EXPORT_SYMBOL(init_wl_funp);

VOID init_hnat_funp(INT (*add_fun)(struct hwnat_tuple *opt), INT (*del_fun)(struct hwnat_tuple *opt), INT (*get_fun)(INT state))
{
	add_hwnat=add_fun;
	del_hwnat=del_fun;
	get_hwnat_entrynum=get_fun;
}
EXPORT_SYMBOL(init_hnat_funp);

static INT get_HNAT_param(struct sk_buff *pskb,
      		                const struct net_device *in,
      		                const struct net_device *out,
      		                struct xt_td_target_info *info,
      		                struct lan_pc_info *pc_info,
      		                struct hnat_param *p_param)
{
	struct hnat_param param={{0}};
	struct iphdr *iph=pskb->nh.iph;
	U_CHAR *s_mac=eth_hdr(pskb)->h_source;
	__BE32 s_ip=iph->saddr;
	__BE32 d_ip=iph->daddr;
	struct net_device *wl_dev=NULL;
	
	memcpy(param.dev_wan_mac, info->wan_mac, MAC_LEN);
	memcpy(param.dev_lan_mac, info->lan_mac, MAC_LEN);
	param.dev_wan_ip=ntohl(info->wan_ip);

	if(pc_info->d_dir==WAN2LAN)
	{
		if(!strcmp(info->wan_mode, "pppoe"))
			memcpy(param.external_mac, info->remote_mac, MAC_LEN);
		else
			memcpy(param.external_mac, s_mac, MAC_LEN);
		if(get_mac_in_arp(out, d_ip, param.internal_mac)!=0)
			return -1;
		param.external_ip=ntohl(s_ip);
		param.internal_ip=ntohl(d_ip);
		if(IPPROTO_TCP==iph->protocol)
		{
			struct tcphdr *tcph=(struct tcphdr *)((CHAR *)iph+(iph->ihl<<2));
			param.external_port=ntohs(tcph->source);
			param.internal_port=ntohs(tcph->dest);
		}
		else if(IPPROTO_UDP==iph->protocol)
		{
			struct udphdr *udph=(struct udphdr *)((CHAR *)iph+(iph->ihl<<2));
			param.external_port=ntohs(udph->source);
			param.internal_port=ntohs(udph->dest);
		}
		else
			return -1;
	}
	else
	{
		if(!strcmp(info->wan_mode, "pppoe"))
			memcpy(param.external_mac, info->remote_mac, MAC_LEN);
		else
		{
			if(get_mac_in_arp(out, d_ip, param.external_mac)!=0)
			{
				//int ip_route_input(struct sk_buff *skb, __be32 daddr, __be32 saddr,u8 tos, struct net_device *dev)
				struct sk_buff *skb2=skb_clone(pskb, GFP_ATOMIC);
				if(!skb2)
					return -1;
				if(ip_route_input(skb2, d_ip, s_ip, iph->tos, skb2->dev)!=0)
				{
					kfree_skb(skb2);
					return -1;
				}
				else
				{
					struct rtable *rt=(struct rtable *)skb2->dst;
					if(get_mac_in_arp(out, rt->rt_gateway, param.external_mac)!=0)
					{
						kfree_skb(skb2);
						return -1;
					}
				}
				X_DBG("%s,%d: rt_gateway<%u.%u.%u.%u>\n", __FUNCTION__, __LINE__, NIPQUAD(((struct rtable *)(skb2->dst))->rt_gateway));
				kfree_skb(skb2);
			}
		}
		memcpy(param.internal_mac, s_mac, MAC_LEN);
		param.external_ip=ntohl(d_ip);
		param.internal_ip=ntohl(s_ip);
		if(IPPROTO_TCP==iph->protocol)
		{
			struct tcphdr *tcph=(struct tcphdr *)((CHAR *)iph+(iph->ihl<<2));
			param.external_port=ntohs(tcph->dest);
			param.internal_port=ntohs(tcph->source);
		}
		else if(IPPROTO_UDP==iph->protocol)
		{
			struct udphdr *udph=(struct udphdr *)((CHAR *)iph+(iph->ihl<<2));
			param.external_port=ntohs(udph->dest);
			param.internal_port=ntohs(udph->source);
		}
		else
			return -1;
	}
	param.dev_wan_port=param.internal_port;
	param.protocol=iph->protocol;
	param.wan_vlan_act=2; //Ins
	param.wan_vlan_id=2;
	param.wan_dst_port=1; //GE1
	param.wan_act_dp=0;
	param.lan_vlan_act=2;
	param.lan_vlan_id=1;
	param.lan_dst_port=1;
	param.lan_act_dp=0;
	if((wl_dev=dev_get_by_name(info->wl_if))!=NULL)
	{
		if(is_wl_client!=NULL && is_wl_client(wl_dev, param.internal_mac)==1)
		{
			X_DBG("%s,%d: is wireless client.\n", __FUNCTION__, __LINE__);
			param.wan_dst_port=2;
			param.wan_act_dp=1;
			param.lan_vlan_act=3; //Del
			param.lan_vlan_id=0;
			param.lan_dst_port=0; //CPU
			param.lan_act_dp=2;
		}
		dev_put(wl_dev);
	}
	if(!strcmp(info->wan_mode, "pppoe"))
	{
		param.lan_pppoe_act=3; //Del
		param.lan_pppoe_id=0;
		param.wan_pppoe_act=2; //Ins
		param.wan_pppoe_id=ntohs(info->pppoe_id);
	}
	else
	{
		param.wan_pppoe_act=param.lan_pppoe_act=0; //No
		param.wan_pppoe_id=param.lan_pppoe_id=0;
	}
	
	X_DBG("%s,%d: hw_nat parameter:\n"
			"\t dev_wan_mac<%02X:%02X:%02X:%02X:%02X:%02X>\n"
			"\t dev_lan_mac<%02X:%02X:%02X:%02X:%02X:%02X>\n"
			"\t external_mac<%02X:%02X:%02X:%02X:%02X:%02X>\n"
			"\t internal_mac<%02X:%02X:%02X:%02X:%02X:%02X>\n"
			"\t dev_wan_ip<%u.%u.%u.%u> external_ip<%u.%u.%u.%u> internal_ip<%u.%u.%u.%u>\n"
			"\t dev_wan_port<%u> external_port<%u> internal_port<%u>\n"
			"\t protocol<%s>\n",
			__FUNCTION__, __LINE__,
			param.dev_wan_mac[0], param.dev_wan_mac[1], param.dev_wan_mac[2], param.dev_wan_mac[3], param.dev_wan_mac[4], param.dev_wan_mac[5],
			param.dev_lan_mac[0], param.dev_lan_mac[1], param.dev_lan_mac[2], param.dev_lan_mac[3], param.dev_lan_mac[4], param.dev_lan_mac[5],
			param.external_mac[0], param.external_mac[1], param.external_mac[2], param.external_mac[3], param.external_mac[4], param.external_mac[5],
			param.internal_mac[0], param.internal_mac[1], param.internal_mac[2], param.internal_mac[3], param.internal_mac[4], param.internal_mac[5],
			NIPQUAD(param.dev_wan_ip), NIPQUAD(param.external_ip), NIPQUAD(param.internal_ip),
			param.dev_wan_port, param.external_port, param.internal_port,
			(IPPROTO_TCP==param.protocol)?"TCP":((IPPROTO_UDP==param.protocol)?"UDP":"OTHER"));

	memcpy(p_param, &param, sizeof(param));
	return 0;
}

static INT add_HNAT(struct sk_buff *pskb,
		                const struct net_device *in,
		                const struct net_device *out,
		                struct xt_td_target_info *info,
		                struct lan_pc_info *pc_info)
{
	struct hnat_param param={{0}};
	struct hwnat_tuple lan_opt={0};
	struct hwnat_tuple wan_opt={0};
	//struct ac_args ac_opt={{0}};

	if(!add_hwnat)
	{
		X_DBG("%s,%d: add_hwnat is NULL\n", __FUNCTION__, __LINE__);
		return -1;
	}
	
	if(get_HNAT_param(pskb, in, out, info, pc_info, &param)!=0)
	{
		PRINT("%s,%d: get_HNAT_param() fail \n", __FUNCTION__, __LINE__);
		return -1;
	}
	
	//hw_nat -a -h [SMAC] -i [DMAC] -j [Sip] -k [Dip] -l [Sp] -m [Dp] -n [New_Sip] -o [New_Dip] -p [New_Sp] -q [New_Dp] -r [Vlan:No/Mod/Ins/Del] -s [VLAN_ID] -t [PPPoE:No/Mod/Ins/Del] -u [PPPoE_ID] -v [Tcp/Udp] -w [OutIf :CPU/GE1/GE2]
	/* lan -> wan options */
	memcpy(lan_opt.smac, param.dev_wan_mac, MAC_LEN);
	memcpy(lan_opt.dmac, param.external_mac, MAC_LEN);
	lan_opt.sip=param.internal_ip;
	lan_opt.dip=param.external_ip;
	lan_opt.sport=param.internal_port;
	lan_opt.dport=param.external_port;
	lan_opt.new_sip=param.dev_wan_ip;
	lan_opt.new_dip=param.external_ip;
	lan_opt.new_sport=param.dev_wan_port;
	lan_opt.new_dport=param.external_port;
	lan_opt.vlan1_act=param.wan_vlan_act;
	lan_opt.vlan1=param.wan_vlan_id;
	lan_opt.pppoe_act=param.wan_pppoe_act;
	lan_opt.pppoe_id=param.wan_pppoe_id;
	lan_opt.is_udp=(param.protocol==IPPROTO_UDP)?1:0;
	lan_opt.dst_port=param.wan_dst_port;
	lan_opt.act_dp=param.wan_act_dp;
	/*wan -> lan options*/
	memcpy(wan_opt.smac, param.dev_lan_mac, MAC_LEN);
	memcpy(wan_opt.dmac, param.internal_mac, MAC_LEN);
	wan_opt.sip=param.external_ip;
	wan_opt.dip=param.dev_wan_ip;
	wan_opt.sport=param.external_port;
	wan_opt.dport=param.dev_wan_port;
	wan_opt.new_sip=param.external_ip;
	wan_opt.new_dip=param.internal_ip;
	wan_opt.new_sport=param.external_port;
	wan_opt.new_dport=param.internal_port;
	wan_opt.vlan1_act=param.lan_vlan_act; //Ins
	wan_opt.vlan1=param.lan_vlan_id;
	wan_opt.pppoe_act=param.lan_pppoe_act; //Del
	wan_opt.pppoe_id=param.lan_pppoe_id;
	wan_opt.is_udp=(param.protocol==IPPROTO_UDP)?1:0;
	wan_opt.dst_port=param.lan_dst_port;
	wan_opt.act_dp=param.lan_act_dp;

	/* ac_opt */
	//ac_opt.ip_s=ac_opt.ip_e=param.external_ip;
	
	if(add_hwnat(&lan_opt)!=0
		|| add_hwnat(&wan_opt)!=0)
	{
		X_DBG("%s,%d: do add_hwnat() fail\n", __FUNCTION__, __LINE__);
		return -1;
	}
	
	return 0;
}

static INT del_HNAT(struct ip_conntrack *ct, struct xt_td_target_info *info)
{
	enum dev_dir d_dir=MAX_DEV_DIR;
	struct hnat_param param={{0}};
	struct hwnat_tuple lan_opt={0};
	struct hwnat_tuple wan_opt={0};

	if(!del_hwnat || !ct || info->lan_subnet[NET_LAN]==0)
	{
		X_DBG("%s,%d: del_hwnat<%s> ct<%s>\n",
				__FUNCTION__, __LINE__,
				del_hwnat?"NOT NULL":"NULL", ct?"NOT NULL":"NULL");
		return -1;
	}

	if((ntohl(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip) & info->lan_subnet[MASK_LAN]) == info->lan_subnet[NET_LAN])
		d_dir=LAN2WAN;
	else if((ntohl(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip) & info->lan_subnet[MASK_LAN]) == info->lan_subnet[NET_LAN])
		d_dir=WAN2LAN;
	else
		return -1;

	if(d_dir==WAN2LAN)
	{
		param.external_ip=ntohl(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip);
		param.internal_ip=ntohl(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip);
		if(IPPROTO_UDP==ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum)
		{
			param.external_port=ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port);
			param.internal_port=ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.udp.port);
		}
		else if(IPPROTO_TCP==ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum)
		{
			param.external_port=ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.tcp.port);
			param.internal_port=ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.tcp.port);
		}
	}
	else
	{
		param.external_ip=ntohl(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip);
		param.internal_ip=ntohl(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip);
		if(IPPROTO_UDP==ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum)
		{
			param.external_port=ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.udp.port);
			param.internal_port=ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port);
		}
		else if(IPPROTO_TCP==ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum)
		{
			param.external_port=ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.tcp.port);
			param.internal_port=ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.tcp.port);
		}
	}
	get_devip_by_name(info->wan_if, &(param.dev_wan_ip));
	param.dev_wan_ip=ntohl(param.dev_wan_ip);
	param.protocol=ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum;
	param.dev_wan_port=param.internal_port;
	X_DBG("%s,%d: hw_nat parameter:\n"
			"\t dev_wan_ip<%u.%u.%u.%u> external_ip<%u.%u.%u.%u> internal_ip<%u.%u.%u.%u>\n"
			"\t dev_wan_port<%u> external_port<%u> internal_port<%u>\n"
			"\t protocol<%s>\n",
			__FUNCTION__, __LINE__,
			NIPQUAD(param.dev_wan_ip), NIPQUAD(param.external_ip), NIPQUAD(param.internal_ip),
			param.dev_wan_port, param.external_port, param.internal_port,
			(IPPROTO_TCP==param.protocol)?"TCP":((IPPROTO_UDP==param.protocol)?"UDP":"OTHER"));
	//lan opt
	lan_opt.sip=param.internal_ip;
	lan_opt.dip=param.external_ip;
	lan_opt.sport=param.internal_port;
	lan_opt.dport=param.external_port;
	lan_opt.is_udp=(param.protocol==IPPROTO_UDP)?1:0;
	//wan opt
	wan_opt.sip=param.external_ip;
	wan_opt.dip=param.dev_wan_ip;
	wan_opt.sport=param.external_port;
	wan_opt.dport=param.dev_wan_port;
	wan_opt.is_udp=(param.protocol==IPPROTO_UDP)?1:0;
	
	del_hwnat(&lan_opt);
	del_hwnat(&wan_opt);
	return 1;
}

VOID destroy_track_record(struct ip_conntrack *ct)
{
	struct lan_pc_info pc_info={0};
	struct ip_conntrack_tuple tuple;
	if(ct->trk_rcd==NULL)
		return;
	tuple.src.ip=ntohl(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip);
	tuple.dst.ip=ntohl(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip);
	tuple.dst.protonum=ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum;
	if(extract_pc_info(&tuple, &td_info, &pc_info)!=-1)
		update_pc_alive_s(&(pc_rcd[pc_info.index]), ct->trk_rcd, ct->trk_rcd->type, -1);
	kfree(ct->trk_rcd);
#if 0
    /* del HNAT */
	if(td_info.hwnat_en==1)
		del_HNAT(ct, &td_info);
#endif
	X_DBG("destroy track_record<%p> for conntrack<%p> pc_info={%d, %d}\n\n",
			ct->trk_rcd, ct, pc_info.index, pc_info.d_dir);
}

static enum session_type
DNS_detect(struct sk_buff *pskb, struct ip_conntrack_tuple *tuple)
{
	if((pskb->len+pskb->mac_len)<MEDIUM_PACKET_LEN
		&& ((tuple->dst.protonum==IPPROTO_TCP
				&& (tuple->src.u.tcp.port==53 || tuple->dst.u.tcp.port==53))
			|| (tuple->dst.protonum==IPPROTO_UDP
				&& (tuple->src.u.udp.port==53 || tuple->dst.u.udp.port==53))))
		return D_DNS_SESSION;
	return MAX_SESSION;
}

static enum session_type
MAIL_detect(struct sk_buff *pskb, struct ip_conntrack_tuple *tuple)
{
	if(tuple->dst.protonum==IPPROTO_TCP
		&& (tuple->src.u.tcp.port==25
			|| tuple->src.u.tcp.port==110
			|| tuple->src.u.tcp.port==143
			|| tuple->dst.u.tcp.port==25
			|| tuple->dst.u.tcp.port==110
			|| tuple->dst.u.tcp.port==143))
		return D_MAIL_SESSION;
	return MAX_SESSION;
}

static enum session_type
FTP_detect(struct sk_buff *pskb, struct ip_conntrack_tuple *tuple)
{
	if(tuple->dst.protonum==IPPROTO_TCP
		&& (tuple->src.u.tcp.port==21
			|| tuple->src.u.tcp.port==20
			|| tuple->dst.u.tcp.port==21
			|| tuple->dst.u.tcp.port==20))
		return D_FTP_SESSION;
	return MAX_SESSION;
}

static enum session_type
HTTP_detect(struct sk_buff *pskb, struct ip_conntrack_tuple *tuple)
{
	struct iphdr *iph=pskb->nh.iph;
	struct tcphdr *tcph=NULL;
	CHAR *data=NULL;
	INT i=0;
	CHAR *methods[]={"GET", "HEAD", "POST", "PUT", "DELETE",
				           "SECURE", "LINK", "UNLINK",
				           "SUBSCRIBE", "UNSUBSCRIBE", 0};
	if(tuple->dst.protonum==IPPROTO_TCP
		&& ((tuple->src.u.tcp.port==80 && tuple->dst.u.tcp.port<HTTP_MAX_PORT)
			|| (tuple->dst.u.tcp.port==80 && tuple->src.u.tcp.port<HTTP_MAX_PORT)))
	{
		tcph=(struct tcphdr *)((CHAR *)iph+(iph->ihl<<2));
		data=(CHAR *)iph+(iph->ihl<<2)+(tcph->doff<<2);
		X_DBG("%s,%d: data<%c%c%c%c>\n", __FUNCTION__, __LINE__, data[0], data[1], data[2], data[3]);
		while(methods[i])
		{
			if(!strncmp(data, methods[i], strlen(methods[i])))
			{
				if(i==0/*!strcmp(methods[i], "GET")*/)
				{
					if(strstr(data, "Range:"))
					{
						X_DBG("%s,%d: return D_BAD_HTTP_SESSION\n", __FUNCTION__, __LINE__);
						return D_BAD_HTTP_SESSION;
					}
				}
				return HTTP_SESSION;
			}
			i++;
		}
	}
	return MAX_SESSION;
}

static enum session_type
special_detect(struct sk_buff *pskb,
		           struct ip_conntrack_tuple *tuple,
		           struct xt_td_target_info *info)
{
	enum session_type cur_type=MAX_SESSION;

	X_DBG("%s,%d: %d <%lx -> %lx>\n",
			__FUNCTION__, __LINE__,
			tuple->dst.protonum, tuple->src.ip, tuple->dst.ip);

	if((cur_type=DNS_detect(pskb, tuple))!=MAX_SESSION)
		return cur_type;
	else if((cur_type=MAIL_detect(pskb, tuple))!=MAX_SESSION)
		return cur_type;
	else if((cur_type=FTP_detect(pskb, tuple))!=MAX_SESSION)
		return cur_type;
	else if((cur_type=HTTP_detect(pskb, tuple))!=MAX_SESSION)
		return cur_type;
	/*MAYBE_GOOD detect*/
	else if((tuple->dst.protonum==IPPROTO_TCP
			&& ((REPLY_DIR==tuple->dst.dir && tuple->src.u.tcp.port<=info->maybe_good && tuple->src.u.tcp.port!=80)
				|| (ORIGINAL_DIR==tuple->dst.dir && tuple->dst.u.tcp.port<=info->maybe_good && tuple->dst.u.tcp.port!=80)))
		|| (tuple->dst.protonum==IPPROTO_UDP
			&& ((tuple->src.u.udp.port<=info->maybe_good && REPLY_DIR==tuple->dst.dir)
				|| (tuple->dst.u.udp.port<=info->maybe_good && ORIGINAL_DIR==tuple->dst.dir))))
		return MAYBE_GOOD_SESSION;
	return MAX_SESSION;
}

static U_INT select_max_noises(struct track_record *trk_rcd,
		                     struct xt_td_target_info *info)
{
	INT i, j, elim_num;
	U_INT tmp_noises=0;

	elim_num=(trk_rcd->small_pkts[ORIGINAL_DIR]+trk_rcd->small_pkts[REPLY_DIR]
			+trk_rcd->medium_pkts[ORIGINAL_DIR]+trk_rcd->medium_pkts[REPLY_DIR]
			+trk_rcd->big_pkts[ORIGINAL_DIR]+trk_rcd->big_pkts[REPLY_DIR])>>info->eliminate_noises;
	X_DBG("%s,%d: pkts={{%u, %u}, {%u, %u}, {%u, %u}} elim_num<%d>\n",
			__FUNCTION__, __LINE__,
			trk_rcd->small_pkts[ORIGINAL_DIR], trk_rcd->small_pkts[REPLY_DIR],
			trk_rcd->medium_pkts[ORIGINAL_DIR], trk_rcd->medium_pkts[REPLY_DIR],
			trk_rcd->big_pkts[ORIGINAL_DIR], trk_rcd->big_pkts[REPLY_DIR], elim_num);
	if(elim_num>=NOISES_NUM)
		return 0;
	for(i=0;i<=elim_num;i++)
	{
		for(j=i;j<NOISES_NUM-1;j++)
		{
			if(trk_rcd->max_intvl[i]<trk_rcd->max_intvl[j+1])
			{
				tmp_noises=trk_rcd->max_intvl[i];
				trk_rcd->max_intvl[i]=trk_rcd->max_intvl[j+1];
				trk_rcd->max_intvl[j+1]=tmp_noises;
			}
		}
	}

	X_DBG("%s, %d: return %u\n", __FUNCTION__, __LINE__, trk_rcd->max_intvl[elim_num]);
	return trk_rcd->max_intvl[elim_num];
}

static enum session_type
VOIP_detect(struct track_record *trk_rcd, struct xt_td_target_info *info)
{
	U_INT max_intvl=select_max_noises(trk_rcd, info);

	X_DBG("%s,%d: intvl<%u : %u> medium_pkts={%u, %u} big_pkts={%u, %u} scale<%d : %d>\n",
			__FUNCTION__, __LINE__, max_intvl, VOIP_MAX_INTERVAL,
			trk_rcd->medium_pkts[ORIGINAL_DIR], trk_rcd->medium_pkts[REPLY_DIR],
			trk_rcd->big_pkts[ORIGINAL_DIR], trk_rcd->big_pkts[REPLY_DIR],
			trk_rcd->small_pkts[ORIGINAL_DIR]+trk_rcd->small_pkts[REPLY_DIR], VOIP_PKTS_LIMIT);

	if(max_intvl<VOIP_MAX_INTERVAL
		&& trk_rcd->medium_pkts[ORIGINAL_DIR]==0
		&& trk_rcd->medium_pkts[REPLY_DIR]==0
		&& trk_rcd->big_pkts[ORIGINAL_DIR]==0
		&& trk_rcd->big_pkts[REPLY_DIR]==0
		&& (trk_rcd->small_pkts[ORIGINAL_DIR]+trk_rcd->small_pkts[REPLY_DIR])>VOIP_PKTS_LIMIT)
		return VOIP_SESSION;

	return MAX_SESSION;
}

static enum session_type
GAMING_detect(struct ip_conntrack_tuple *tuple,
		          struct track_record *trk_rcd,
		          struct xt_td_target_info *info)
{
	U_INT max_intvl=select_max_noises(trk_rcd, info);
	U_INT tot_pkts=trk_rcd->small_pkts[ORIGINAL_DIR]
			          +trk_rcd->small_pkts[REPLY_DIR]
			          +trk_rcd->medium_pkts[ORIGINAL_DIR]
			          +trk_rcd->medium_pkts[REPLY_DIR]
			          +trk_rcd->big_pkts[ORIGINAL_DIR]
			          +trk_rcd->big_pkts[REPLY_DIR];
	U_INT big_pkts=trk_rcd->big_pkts[ORIGINAL_DIR]
			          +trk_rcd->big_pkts[REPLY_DIR];
	U_INT o_pkts=trk_rcd->small_pkts[ORIGINAL_DIR]
			        +trk_rcd->medium_pkts[ORIGINAL_DIR]
			        +trk_rcd->big_pkts[ORIGINAL_DIR];
	U_INT r_pkts=trk_rcd->small_pkts[REPLY_DIR]
			        +trk_rcd->medium_pkts[REPLY_DIR]
			        +trk_rcd->big_pkts[REPLY_DIR];

	X_DBG("%s,%d: intvl[%u : %u] big_pkts[%u : %u] tot_pkts[%u : %u] traffic[%lu : %lu] o_r[%u : %u]\n",
			__FUNCTION__, __LINE__,
			max_intvl, GAMING_MAX_INTERVAL,
			(big_pkts<<GAMING_BIGPKTS_SCALE), tot_pkts,
			tot_pkts, GAMING_LEAST_PKTS,
			trk_rcd->traffic, GAMING_MAX_TRAFFIC,
			o_pkts, r_pkts);

	if(max_intvl<GAMING_MAX_INTERVAL
		&& (big_pkts<<GAMING_BIGPKTS_SCALE)<tot_pkts
		&& tot_pkts>=GAMING_LEAST_PKTS
		&& trk_rcd->traffic<GAMING_MAX_TRAFFIC
		&& ((o_pkts<=r_pkts && (o_pkts<<1)>r_pkts) || (o_pkts>=r_pkts && o_pkts<(r_pkts<<1))))
		return GAMING_SESSION;

	return MAX_SESSION;
}

static enum session_type
P2P_detect(struct track_record *trk_rcd,
		       struct pc_record *p_pcrcd,
		       struct xt_td_target_info *info)
{
	INT bad_session=p_pcrcd->alive_s[FIRST_PKT_SESSION]
			           +p_pcrcd->alive_s[UNSAFE_SESSION]
			           +p_pcrcd->alive_s[P2P_SESSION]
			           +p_pcrcd->alive_s[D_P2P_SESSION];

	X_DBG("%s,%d: big_pkts={%u, %u} session<%d+%d+%d+%d : %d>\n",
			__FUNCTION__, __LINE__,
			trk_rcd->big_pkts[ORIGINAL_DIR], trk_rcd->big_pkts[REPLY_DIR],
			p_pcrcd->alive_s[FIRST_PKT_SESSION], p_pcrcd->alive_s[UNSAFE_SESSION],
			p_pcrcd->alive_s[P2P_SESSION], p_pcrcd->alive_s[D_P2P_SESSION], P2P_MAX_BAD_SESSIONS);

	if(((trk_rcd->big_pkts[ORIGINAL_DIR]>trk_rcd->big_pkts[REPLY_DIR]
				&& trk_rcd->big_pkts[REPLY_DIR]<P2P_PKTS_LIMIT)
			|| (trk_rcd->big_pkts[ORIGINAL_DIR]<trk_rcd->big_pkts[REPLY_DIR]
				&& trk_rcd->big_pkts[ORIGINAL_DIR]<P2P_PKTS_LIMIT))
		|| bad_session>P2P_MAX_BAD_SESSIONS)
		return P2P_SESSION;

	return MAX_SESSION;
}

static enum session_type
common_detect(struct sk_buff *pskb,
		          struct ip_conntrack_tuple *tuple,
		          struct track_record *trk_rcd,
		          struct pc_record *p_pcrcd,
		          struct xt_td_target_info *info)
{
	enum session_type cur_type=MAX_SESSION;

	if(trk_rcd->type==HTTP_SESSION)
	{
		if(relative_tenthsecs(jiffies, trk_rcd->initial_time)>info->long_life_time)
			cur_type=D_BAD_HTTP_SESSION;
		else
			cur_type=HTTP_SESSION;
	}
	else if((cur_type=VOIP_detect(trk_rcd, info))!=MAX_SESSION)
		goto check_type;
	else if((cur_type=GAMING_detect(tuple, trk_rcd, info))!=MAX_SESSION)
		goto check_type;
	else if((cur_type=P2P_detect(trk_rcd, p_pcrcd, info))!=MAX_SESSION)
		goto check_type;
	else
	{
		if(relative_tenthsecs(jiffies, trk_rcd->initial_time)>info->long_life_time
			&& trk_rcd->traffic>info->large_traffic)
			cur_type=BAD_UNKNOWN_SESSION;
		else
			cur_type=UNKNOWN_SESSION;
	}

check_type:
	if(cur_type==trk_rcd->type)
	{
		switch(cur_type)
		{
			case UNKNOWN_SESSION:
				cur_type=D_UNKNOWN_SESSION;
				break;
			case BAD_UNKNOWN_SESSION:
				cur_type=D_BAD_UNKNOWN_SESSION;
				break;
			case VOIP_SESSION:
				cur_type=D_VOIP_SESSION;
				break;
			case GAMING_SESSION:
				cur_type=D_GAMING_SESSION;
				break;
			case P2P_SESSION:
				cur_type=D_P2P_SESSION;
				break;
			default:
				break;
		}
	}

	/*clean some record*/
	trk_rcd->traffic=0;
	memset(trk_rcd->max_intvl, 0, sizeof(trk_rcd->max_intvl));
	memset(trk_rcd->small_pkts, 0, sizeof(trk_rcd->small_pkts));
	memset(trk_rcd->medium_pkts, 0, sizeof(trk_rcd->medium_pkts));
	memset(trk_rcd->big_pkts, 0, sizeof(trk_rcd->big_pkts));
	
	X_DBG("%s,%d: return cur_type<%d>\n", __FUNCTION__, __LINE__, cur_type);
	return cur_type;
}

static enum session_type
do_statistic(struct sk_buff *pskb,
		         struct ip_conntrack_tuple *tuple,
		         struct track_record *trk_rcd,
		         struct xt_td_target_info *info,
		         struct pc_record *p_pcrcd,
		         U_INT intvl)
{
	enum session_type cur_type=MAX_SESSION;
	INT min_in_max, i;
	U_INT pkts_len=pskb->len+pskb->mac_len;

	X_DBG("%s,%d: %d <%lx -> %lx>\n",
			__FUNCTION__, __LINE__,
			tuple->dst.protonum, tuple->src.ip, tuple->dst.ip);

	if(pkts_len<SMALL_PACKET_LEN)
		trk_rcd->small_pkts[tuple->dst.dir]++;
	else if(pkts_len>BIG_PACKET_LEN)
		trk_rcd->big_pkts[tuple->dst.dir]++;
	else
		trk_rcd->medium_pkts[tuple->dst.dir]++;

	trk_rcd->traffic+=pkts_len;

	if(trk_rcd->time_stamp!=-1)
	{
		min_in_max=0;
		for(i=0;i<NOISES_NUM;i++)
		{
			if(trk_rcd->max_intvl[i]==0)
			{
				min_in_max=i;
				break;
			}
			else /* find minimum in max_intvl */
				min_in_max=(trk_rcd->max_intvl[i]<trk_rcd->max_intvl[min_in_max])?i:min_in_max;
		}
		trk_rcd->max_intvl[min_in_max]=intvl;

		if(trk_rcd->type==FIRST_PKT_SESSION)
			cur_type=EARLY_UNKNOWN_SESSION;
		if(p_pcrcd->s_limit>0 && trk_rcd->type<BAD_UNKNOWN_SESSION)
			cur_type=UNSAFE_SESSION;
	}
	X_DBG("%s,%d: return cur_type<%d>\n", __FUNCTION__, __LINE__, cur_type);
	return cur_type;
}

static INT
TD_check(const CHAR *tablename,
	       const VOID *e,
	       const struct xt_target *target,
	       VOID *targinfo,
	       U_INT hook_mask)
{
	struct xt_td_target_info *info=(struct xt_td_target_info *)targinfo;

	X_DBG("%s,%d: sessions<%u> downstream<%lu> upstream<%lu> wan_mode<%s> wan_if<%s> lan_if<%s> lan_subnet<%lx/%lx>\n",
			__FUNCTION__, __LINE__,
			info->sessions, info->downstream, info->upstream,
			info->wan_mode, info->wan_if, info->lan_if, info->lan_subnet[NET_LAN], info->lan_subnet[MASK_LAN]);

	if(info->sessions==0 || info->downstream==0 || info->upstream==0
		|| info->wan_mode[0]=='\0' || info->wan_if[0]=='\0' || info->lan_if[0]=='\0'
		|| info->lan_subnet[NET_LAN]==0 || info->lan_subnet[MASK_LAN]==0)
	{
		X_DBG("startup TrafficDetection fail\n");
		return 0;
	}

	memcpy(&td_info, info, sizeof(struct xt_td_target_info));
	if(!strcmp(td_info.wan_mode, "pppoe"))
	{
		read_lock(&pppox_lock);
		get_devmac_by_name(s_pppox.sa_addr.pppoe.dev, td_info.wan_mac);
		memcpy(td_info.remote_mac, s_pppox.sa_addr.pppoe.remote, MAC_LEN);
		td_info.pppoe_id=s_pppox.sa_addr.pppoe.sid;
		read_unlock(&pppox_lock);
	}
	else
		get_devmac_by_name(td_info.wan_if, td_info.wan_mac);
	get_devmac_by_name(td_info.lan_if, td_info.lan_mac);
	get_devip_by_name(td_info.wan_if, &(td_info.wan_ip));
	get_devip_by_name(td_info.lan_if, &(td_info.lan_ip));
	
	td_info.track_time=(td_info.track_time>0)?td_info.track_time:TRACK_TIME;
	td_info.sleep_time=(td_info.sleep_time>0)?td_info.sleep_time:SLEEP_TIME;
	td_info.long_sleep_time=(td_info.long_sleep_time>0)?td_info.long_sleep_time:LONG_SLEEP_TIME;
	td_info.tcp_track_delay=(td_info.tcp_track_delay>0)?td_info.tcp_track_delay:TCP_TRACK_DELAY;
	td_info.long_life_time=(td_info.long_life_time>0)?td_info.long_life_time:LONG_LIFE_TIME;
	td_info.large_traffic=(td_info.large_traffic>0)?td_info.large_traffic:LARGE_TRAFFIC;
	td_info.maybe_good=(td_info.maybe_good>0)?td_info.maybe_good:MAYBE_GOOD_PORT_LIMIT;
	td_info.eliminate_noises=(td_info.eliminate_noises>0)?td_info.eliminate_noises:ELIMINATE_NOISES_SCALE;

	X_DBG("%s,%d: tablename<%s> hook_mask<%d>\n"
			"\t track_time<%u> sleep_time<%u> long_sleep_time<%u>\n"
			"\t tcp_track_delay<%u> long_life_time<%u> large_traffic<%lu>\n"
			"\t maybe_good<%u> eliminate_noises<%u>\n"
			__FUNCTION__, __LINE__, tablename, hook_mask,
			td_info.track_time, td_info.sleep_time, td_info.long_sleep_time,
			td_info.tcp_track_delay, td_info.long_life_time, td_info.large_traffic,
			td_info.maybe_good, td_info.eliminate_noises);

	write_lock(&startup_lock);
	startup.time_stamp=jiffies;
	startup.reset=1;
	startup.fore_stat=1;
	write_unlock(&startup_lock);
	
	write_lock(&stat_lock);
	memset(stat_tab, 0, sizeof(stat_tab));
	write_unlock(&stat_lock);

	memset(pc_rcd, 0, sizeof(pc_rcd));

	return 1;
}

static U_INT
TD_target(struct sk_buff **pskb,
		const struct net_device *in,
		const struct net_device *out,
		U_INT hooknum,
		const struct xt_target *target,
		const VOID *targinfo)
{
	struct xt_td_target_info *info=&td_info;
	enum ip_conntrack_info ctinfo;
	struct ip_conntrack *ct;
	struct ip_conntrack_tuple tuple;
	struct track_record *trk_rcd=NULL;
	struct iphdr *iph=(*pskb)->nh.iph;
	enum session_type cur_type=MAX_SESSION;
	struct lan_pc_info pc_info={0};
	U_INT cur_time;

	X_DBG("%s,%d: hooknum<%d> iph={%d, %d, %lx, %lx} jiffies<%lu>\n",
			__FUNCTION__, __LINE__, hooknum,
			ntohs(iph->tot_len), iph->protocol, ntohl(iph->saddr), ntohl(iph->daddr), jiffies);

    //if (hooknum!=NF_IP_FORWARD)
    //		return XT_CONTINUE;

    if((*pskb)->mark==TYPE2MARK(USRDEF_HW_NAT_PASS))
        return XT_CONTINUE;
    
    if((*pskb)->mark==TYPE2MARK(USRDEF_HW_NAT_UNPASS))
    {
        
        return XT_CONTINUE;
    }
	
    /* get ip_conntrack */
	if(!(ct=ip_conntrack_get(*pskb, &ctinfo)))
		return XT_CONTINUE;

    /* built tuple by sk_buff and convert byte order */
	tuple.src.ip=ntohl(iph->saddr);
	tuple.dst.ip=ntohl(iph->daddr);
	tuple.dst.protonum=iph->protocol;
	tuple.dst.dir=(CTINFO2DIR(ctinfo)==IP_CT_DIR_ORIGINAL)?ORIGINAL_DIR:REPLY_DIR;

	if(extract_pc_info(&tuple, info, &pc_info)==-1)
		return XT_CONTINUE;
	
	/* get track_record of this ip_conntrack */
	if(!ct->trk_rcd)
	{
		ct->trk_rcd=kzalloc(sizeof(struct track_record), GFP_KERNEL);
		if(!(ct->trk_rcd))
		{
			PRINT("Can not allocated track_record for this conntrack<%p>.\n\n", ct);
			return XT_CONTINUE;
		}

		if(relative_tenthsecs(jiffies, pc_rcd[pc_info.index].time_stamp)>TRACK_TIME)
		{
			PRINT("%s,%d: pc_rcd[%d]:::\n", __FUNCTION__, __LINE__, pc_info.index);
			DUMP_PC_RCD((&(pc_rcd[pc_info.index])));
			if(pc_rcd[pc_info.index].term_s<FEWER_PC_SAFE_NEW_SESSIONS && pc_rcd[pc_info.index].s_limit>0)
			{
				pc_rcd[pc_info.index].s_limit++;
				if(pc_rcd[pc_info.index].s_limit>=PC_EXPIATION_TIMES)
					pc_rcd[pc_info.index].s_limit=0;
			}
			pc_rcd[pc_info.index].term_s=0;
			pc_rcd[pc_info.index].time_stamp=jiffies;
		}
		pc_rcd[pc_info.index].term_s++;
		if((pc_rcd[pc_info.index].s_limit>0 && pc_rcd[pc_info.index].term_s>FEWER_PC_SAFE_NEW_SESSIONS)
			|| pc_rcd[pc_info.index].term_s>PC_SAFE_NEW_SESSIONS)
			pc_rcd[pc_info.index].s_limit=1;

		ct->trk_rcd->initial_time=jiffies;
		ct->trk_rcd->time_stamp=-1;
		if(IP_CT_NEW==ctinfo)
		{
			ct->trk_rcd->start_time=-1;
			if(pc_rcd[pc_info.index].s_limit==1)
				ct->trk_rcd->type=UNSAFE_SESSION;
			else
				ct->trk_rcd->type=FIRST_PKT_SESSION;
		}
		else
		{
			ct->trk_rcd->start_time=relative_tenthsecs(jiffies, (ct->trk_rcd)->initial_time);
			if(pc_rcd[pc_info.index].s_limit>0)
				ct->trk_rcd->type=UNSAFE_SESSION;
			else
				ct->trk_rcd->type=EARLY_UNKNOWN_SESSION;
		}
		update_pc_alive_s(&(pc_rcd[pc_info.index]), ct->trk_rcd, ct->trk_rcd->type, 1);
		X_DBG("initialize track_record<%p> for conntrack<%p> pc_info={%d, %d}\n",
				ct->trk_rcd, ct, pc_info.index, pc_info.d_dir);
	}
	trk_rcd=ct->trk_rcd;
	
	/* initialize the relative time */
	cur_time=relative_tenthsecs(jiffies, trk_rcd->initial_time);
	
	/* detection */
	if(trk_rcd->type>D_BAD_UNKNOWN_SESSION)
		goto determine;

	/* extract more info from sk_buff */
	if(IPPROTO_TCP==iph->protocol)
	{
		struct tcphdr *tcph=(struct tcphdr *)((CHAR *)iph+(iph->ihl<<2));
		if(ntohs(iph->tot_len)<=(iph->ihl<<2)+(tcph->doff<<2))
			goto the_end;
		else
		{
			tuple.src.u.tcp.port=ntohs(tcph->source);
			tuple.dst.u.tcp.port=ntohs(tcph->dest);
		}
	}
	else if(IPPROTO_UDP==iph->protocol)
	{
		struct udphdr *udph=(struct udphdr *)((CHAR *)iph+(iph->ihl<<2));
		if(ntohs(iph->tot_len)<=(iph->ihl<<2)+(sizeof(struct udphdr)))
			goto the_end;
		else
		{
			tuple.src.u.udp.port=ntohs(udph->source);
			tuple.dst.u.udp.port=ntohs(udph->dest);
		}
	}
	else
		goto the_end;

	/* do specially detect */
	if(trk_rcd->type<MAYBE_GOOD_SESSION)
	{
		cur_type=special_detect(*pskb, &tuple, info);
		update_pc_alive_s(&(pc_rcd[pc_info.index]), trk_rcd, cur_type, 0);
		if(cur_type<MAX_SESSION && cur_type>D_BAD_UNKNOWN_SESSION)
			goto determine;
	}

	/* is the time to do statistic ? */
	X_DBG("%s,%d: type<%d> time<%u : %u(%lu-%lu)>\n",
			__FUNCTION__, __LINE__, trk_rcd->type,
			trk_rcd->start_time, cur_time, jiffies, trk_rcd->initial_time);
	if(trk_rcd->start_time==-1) /* have not data packet be transfered on this conntrack before */
	{
		if(IPPROTO_TCP==tuple.dst.protonum)
		{
			trk_rcd->start_time=cur_time+info->tcp_track_delay;
			X_DBG("%s,%d: the first TCP data packet, need delay\n", __FUNCTION__, __LINE__);
			goto the_end;
		}
		else
		{
			trk_rcd->start_time=cur_time;
			X_DBG("%s,%d: the first time to do_statistic()\n", __FUNCTION__, __LINE__);
		}
	}
	else
	{
		if(cur_time<trk_rcd->start_time) /* idle time */
		{
			X_DBG("%s,%d: sleeping\n", __FUNCTION__, __LINE__);
			goto the_end;
		}
		else if(cur_time-trk_rcd->start_time>info->track_time) /* track time out */
		{
			if(trk_rcd->traffic>0) /* have not do common_detect() before */
			{
				X_DBG("%s,%d: one statistic finish, do common_detect()\n", __FUNCTION__, __LINE__);
				cur_type=common_detect(*pskb, &tuple, trk_rcd, &(pc_rcd[pc_info.index]), info);
				update_pc_alive_s(&(pc_rcd[pc_info.index]), trk_rcd, cur_type, 0);
				if(cur_type>D_BAD_UNKNOWN_SESSION)
					goto determine;
			}
			if((trk_rcd->type>=D_UNKNOWN_SESSION && cur_time-trk_rcd->start_time>info->track_time+info->long_sleep_time)
				|| (trk_rcd->type<D_UNKNOWN_SESSION && cur_time-trk_rcd->start_time>info->track_time+info->sleep_time))
			{
				X_DBG("%s,%d: will do_statistic() after common_detect()\n", __FUNCTION__, __LINE__);
				trk_rcd->start_time=cur_time;
			}
			else
			{
				if(trk_rcd->type>=D_UNKNOWN_SESSION)
					trk_rcd->start_time+=(info->track_time+info->long_sleep_time);
				else
					trk_rcd->start_time+=(info->track_time+info->sleep_time);
				X_DBG("%s,%d: sleeping after common_detect() <%u : %u>\n", __FUNCTION__, __LINE__, trk_rcd->start_time, cur_time);
				goto the_end;
			}
		}
	}

	/* do statistic */
	cur_type=do_statistic(*pskb, &tuple, trk_rcd, info, &(pc_rcd[pc_info.index]), cur_time-trk_rcd->time_stamp);
	update_pc_alive_s(&(pc_rcd[pc_info.index]), trk_rcd, cur_type, 0);

determine:
#if 0
	/* let this session goto HNAT */
	if( 0 && info->hwnat_en==1)
	{
		if(/*trk_rcd->type==D_DNS_SESSION
			|| */trk_rcd->type==D_VOIP_SESSION
			|| trk_rcd->type==D_GAMING_SESSION)
		add_HNAT(*pskb, in, out, info, &pc_info);
	}
#endif

the_end:
	trk_rcd->time_stamp=cur_time;
	write_lock(&startup_lock);
	if(startup.fore_stat==1)
	{
		if(relative_tenthsecs(jiffies, startup.time_stamp)<=TRACK_TIME)
		{
			U_INT pc_dim_s=pc_rcd[pc_info.index].alive_s[FIRST_PKT_SESSION]
					          +pc_rcd[pc_info.index].alive_s[EARLY_UNKNOWN_SESSION]
					          +pc_rcd[pc_info.index].alive_s[UNSAFE_SESSION];
			pc_rcd[pc_info.index].term_t+=((*pskb)->len+(*pskb)->mac_len);
			X_DBG("%s,%d: pc_rcd[%d] <%lu : %lu> <%u : %u> <%lu/%d>\n",
					__FUNCTION__, __LINE__, pc_info.index,
					pc_rcd[pc_info.index].term_t, PC_MAX_TRAFFIC_GAGE(((info->downstream+info->upstream)*TENTHSEC2SEC(TRACK_TIME))),
					pc_dim_s, PC_MAX_SESSION_GAGE(info->sessions),
					jiffies, HZ);
			if(pc_rcd[pc_info.index].term_t>PC_MAX_TRAFFIC_GAGE(((info->downstream+info->upstream)*TENTHSEC2SEC(TRACK_TIME)))
				|| pc_dim_s>PC_MAX_SESSION_GAGE(info->sessions))
				pc_rcd[pc_info.index].s_limit=1;
		}
		else
			startup.fore_stat=0;
	}
	write_unlock(&startup_lock);
	if(pc_info.d_dir<MAX_DEV_DIR && trk_rcd->type<MAX_SESSION)
	{
		/* statistic for per type session */
		write_lock(&stat_lock);
		stat_tab[pc_info.d_dir].pkt_num[trk_rcd->type]++;
		stat_tab[pc_info.d_dir].traffic[trk_rcd->type]+=((*pskb)->len+(*pskb)->mac_len);
		write_unlock(&stat_lock);
	}
    
    (*pskb)->mark=TYPE2MARK(trk_rcd->type);
	return XT_CONTINUE;
}

static struct xt_target TD_reg = {
	.name       = "TD",
	.target     = TD_target,
	.checkentry = TD_check,
	.me		      = THIS_MODULE,
	.table      = "mangle",
	.targetsize = sizeof(struct xt_td_target_info),
	.hooks      = 1<<NF_IP_FORWARD,
	.family     = AF_INET,
	.revision   = 0,
};

/*
** use proc to get pppoe server's info
*/
static struct proc_dir_entry *pppoe_file;

static ssize_t proc_read_pppoe_fops(struct file *filp, CHAR *buf, size_t count, loff_t *offp)
{
	U_CHAR buffer[QTR_K]={0};
	INT len=0;

	if(*offp!=0)
		return 0;

	read_lock(&pppox_lock);
	len=sprintf(buffer, "remote<%02X:%02X:%02X:%02X:%02X:%02X> sid<%u> dev<%s>\n",
			s_pppox.sa_addr.pppoe.remote[0], s_pppox.sa_addr.pppoe.remote[1], s_pppox.sa_addr.pppoe.remote[2],
			s_pppox.sa_addr.pppoe.remote[3], s_pppox.sa_addr.pppoe.remote[4], s_pppox.sa_addr.pppoe.remote[5],
			ntohs(s_pppox.sa_addr.pppoe.sid), s_pppox.sa_addr.pppoe.dev);

	read_unlock(&pppox_lock);
	len+=1;
	copy_to_user(buf, buffer, len);
	*offp=len;
	return len;
}
static ssize_t proc_write_pppoe_fops(struct file *filp, const CHAR *buffer, size_t count, loff_t *offp)
{
	write_lock(&pppox_lock);
	memset(&s_pppox, 0, sizeof(s_pppox));
	copy_from_user(&s_pppox, buffer, count);
	
	X_DBG("%s,%d: remote<%02X:%02X:%02X:%02X:%02X:%02X> sid<%u> dev<%s>\n",
			__FUNCTION__, __LINE__,
			s_pppox.sa_addr.pppoe.remote[0], s_pppox.sa_addr.pppoe.remote[1], s_pppox.sa_addr.pppoe.remote[2],
			s_pppox.sa_addr.pppoe.remote[3], s_pppox.sa_addr.pppoe.remote[4], s_pppox.sa_addr.pppoe.remote[5],
			s_pppox.sa_addr.pppoe.sid, s_pppox.sa_addr.pppoe.dev);
	
	write_unlock(&pppox_lock);
	return count;
}
static struct file_operations pppoe_fops = {
    read: proc_read_pppoe_fops,
    write: proc_write_pppoe_fops,
};
static INT create_pppoe_proc(VOID)
{
	pppoe_file=create_proc_entry("td_pppoe", 0666, &proc_root);
	pppoe_file->owner=THIS_MODULE;
	pppoe_file->proc_fops=&pppoe_fops;
	return 0;
}
static INT remove_pppoe_proc(VOID)
{
	remove_proc_entry("td_pppoe",&proc_root);
	return 0;
}

/*
** transfer stat_tab between kernel and user space
*/
static struct sock *nl_sk=NULL; /* create netlink */

static VOID netlink_TD_process(struct sock *sk, INT len)
{
	struct sk_buff *skb;
	struct nlmsghdr *nlh=NULL;
	struct last_stat *data=NULL;
	INT ret=-1;
	while((skb=skb_dequeue(&sk->sk_receive_queue))!=NULL)
	{
		/* process netlink message pointed by skb->data */
		nlh=(struct nlmsghdr *)skb->data;
		/* process netlink message with header pointed by nlh and data pointed by data */
		X_DBG("%s,%d: nlh={%d, %d, %d, %d, %d}\n",
				__FUNCTION__, __LINE__,
				nlh->nlmsg_len, nlh->nlmsg_type, nlh->nlmsg_flags, nlh->nlmsg_seq, nlh->nlmsg_pid);
		if(REQUEST_NL_TD==nlh->nlmsg_type)
		{
			data=(struct last_stat *)NLMSG_DATA(nlh);

			write_lock(&stat_lock);
			memcpy(data, stat_tab, sizeof(stat_tab));
			//DUMP_STAT(stat_tab);
			memset(stat_tab, 0, sizeof(stat_tab));
			write_unlock(&stat_lock);

			nlh->nlmsg_type=READY_NL_TD;
			NETLINK_CB(skb).pid=0;        /* from kernel */
			NETLINK_CB(skb).dst_group=0;  /* unicast */
			ret=netlink_unicast(nl_sk, skb, nlh->nlmsg_pid, MSG_DONTWAIT);
			X_DBG("%s,%d: netlink_unicast() end and nlh={%d, %d, %d, %d, %d}\n",
					__FUNCTION__, __LINE__,
					nlh->nlmsg_len, nlh->nlmsg_type, nlh->nlmsg_flags, nlh->nlmsg_seq, nlh->nlmsg_pid);
		}
		else
			kfree_skb(skb);
	}
}

static INT netlink_TD_init(VOID)
{
	nl_sk=netlink_kernel_create(NETLINK_TD, 0, netlink_TD_process, THIS_MODULE);
	if(!nl_sk)
	{
		X_DBG("%s,%d: can not create a netlink socket\n", __FUNCTION__, __LINE__);
		return -1;
	}
	return 0;
}

static VOID netlink_TD_fini(VOID)
{
	sock_release(nl_sk->sk_socket);
}

static INT __init ipt_TD_init(VOID)
{
	INT ret=-1;
	rwlock_init(&stat_lock);
	ret=netlink_TD_init();
	if(ret<0)
		return -1;
	ret=xt_register_target(&TD_reg);
	if(ret<0)
		netlink_TD_fini();
	create_pppoe_proc();
	return ret;
}

static VOID __exit ipt_TD_fini(VOID)
{
	netlink_TD_fini();
	xt_unregister_target(&TD_reg);
	remove_pppoe_proc();
}

module_init(ipt_TD_init);
module_exit(ipt_TD_fini);



#endif /* _SMART_QOS_ */
