/* String matching match for iptables
 *
 * (C) 2005 Pablo Neira Ayuso <pablo@eurodev.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_string.h>
#include <linux/textsearch.h>
#include <linux/tcp.h>
#include <linux/ip.h>
#include <linux/slab.h>

MODULE_AUTHOR("Pablo Neira Ayuso <pablo@eurodev.net>");
MODULE_DESCRIPTION("IP tables string match module");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_string");

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

struct url_list url_info = {NULL, 0};

char* StrIStr( const char* dst , const char* src )
{
	   char* dstp2 = dst;
     char* dstp;
     char* srcp;

     if( dst == NULL
         || src == NULL )
     {
         return NULL;
     }
     while( *dstp2 )
     {
         dstp = dstp2;
         srcp = src;
         while( *dstp
             && *srcp
             && ( ( *dstp | 32 ) == ( *srcp | 32 ) ) )
         {
             ++dstp;
             ++srcp;
         }
         if( !( *srcp ) )
         {
             return dstp2;
         }
         if( !( *dstp ) )
         {
             return NULL;
         }
         ++dstp2;
     }
     return NULL;
}


int get_list_find(struct iphdr *ip,char *temp_find, struct tcphdr *th)
{
	url_node *unode=NULL,*unode1=NULL;
	unode=url_info.head;
	unode1=unode;
	while(url_info.head!=NULL)
	{	
		if(unode->source_port==htons(th->source) && unode->dest_port==htons(th->dest) && unode->seq_get == (htonl(th->seq)-unode->data_len_get))
		{
			strncpy(temp_find, unode->str_uri_get, HOST_STR_SIZE);
			if(unode==url_info.head)
				url_info.head=url_info.head->next;
			else
				unode1->next=unode->next;
			kfree(unode);
			url_info.num--;
			
			return 1;
		}
		unode1=unode;
		unode=unode1->next;
	}
	return 0;
}

void get_list_add(char *temp_get,struct tcphdr *th,struct iphdr *ip)
{
	url_node *unode=NULL,*unode1=NULL;
	if(url_info.head!=NULL)
	{
		unode=url_info.head;
		unode1=unode;
		
		if(url_info.num >= GET_LIST_SIZE)
		{
			url_info.head=url_info.head->next;
			kfree(unode);
			unode=url_info.head;
			unode1=unode;
			url_info.num--;
		}
		while(unode!=NULL)
		{
			unode1=unode;
			unode=unode1->next;
		}
	}
	if(unode=(url_node *)kmalloc(sizeof(url_node),GFP_KERNEL))
		return NULL;
	strncpy(unode->str_uri_get, temp_get, GET_STR_SIZE);
	unode->source_port=htons(th->source);
	unode->dest_port=htons(th->dest);
	unode->seq_get=htonl(th->seq);
	unode->data_len_get=htons(ip->tot_len)-(ip->ihl * 4)-(th->doff * 4);
	unode->next=NULL;
	if(url_info.head==NULL)
		url_info.head=unode;
	else
		unode1->next = unode;
	url_info.num++;
}


static int match(const struct sk_buff *skb,
		 const struct net_device *in,
		 const struct net_device *out,
		 const struct xt_match *match,
		 const void *matchinfo,
		 int offset,
		 unsigned int protoff,
		 int *hotdrop)
{
	const struct xt_string_info *conf = matchinfo;
	struct ts_state state;
	memset(&state, 0, sizeof(struct ts_state));
	
	int filter_key_len,temp_find=0,temp_search=0;	
	char *filter_key, *haystack;
	struct iphdr *ip = skb->nh.iph;
	struct tcphdr *th = (struct tcphdr *)((char *)ip + (ip->ihl << 2));

	haystack = (char *)ip + (ip->ihl << 2) + (th->doff << 2);
	filter_key=(char *)&conf->pattern;	
	filter_key_len = strlen(filter_key);
	temp_find =	search_linear(ip,filter_key, haystack, filter_key_len,th);

	if(temp_find==NULL)
	{
		temp_search=skb_find_text((struct sk_buff *)skb, conf->from_offset,conf->to_offset, conf->config, &state);
		return (temp_search!=UINT_MAX) ^ conf->invert;
	}
	else
	{
			return ((temp_find!=NULL) ^ conf->invert);
	}
}

char *search_linear (struct iphdr *ip,char *filter_key, char *haystack, int filter_key_len,struct tcphdr *th) 
{
	char *t = haystack, *now;
	char temp_search[GET_STR_SIZE+HOST_STR_SIZE]="",temp_get[GET_STR_SIZE]="";
	int len = 0, i = 0;
	if(strnicmp(filter_key, "http://", 7) == 0)
	{
		if(filter_key_len <= 7)return NULL;
		filter_key += 7; // climb the "http://"		
		filter_key_len -= 7;
	}
	t=StrIStr(haystack, "\nhost:");
	if(t)
	{
		t +=6;
		DEBUGP("match line like: Host: www.xxx.com \n");

		now = strchr(t, '\r\n');
		if(!now)
			return NULL;
		DEBUGP("now - t: %d\n", now - t);		
		strncpy(temp_search,t,((((now-t)>HOST_STR_SIZE)? HOST_STR_SIZE:(now-t-1))));
		t = haystack;
		t = strstr(t, "GET ");
		if(t)//have "\nHost: " and "GET "
		{
			DEBUGP("match line like: GET \n");
			t += 4;	// climb "GET "
			now = strchr(t, '\r\n');
			if(!now)
				return NULL;
			now +=2;
			strncpy(temp_get,t,((now-t))>GET_STR_SIZE ? GET_STR_SIZE: (now-t));
			strcat(temp_search,temp_get);
			t=temp_search;
			now = strstr(t, " HTTP");
			if(!now)
				return NULL;
			len = now - t;

			for(i = 0; len - i >= filter_key_len; i++)	
				{
					if(strnicmp(t + i, filter_key, filter_key_len) == 0)
								return 1;
				}
		}
		else//just have "\nHost:" but no "get"
			{	
				char temp_find[HOST_STR_SIZE];
				if(get_list_find(ip,temp_find,th))
					{
						strcat(temp_search,temp_find);
						t=temp_search;
						now = strchr(t, '\r\n');
						if(!now)
							return NULL;
						len = now - t;

						for(i = 0; len - i >= filter_key_len; i++)	
							{
								if(strnicmp(t + i, filter_key, filter_key_len) == 0)
									{
										/* http header was ended by "\r\n\r\n" */
										/* we should can find the end of a  http header */
										now = strstr(haystack, "\r\n\r\n");
										if(now)
											return 1;
									}
							}
					}
					else
						return NULL;
				}
	}
	else //do not have "\nHost:" just have "get"
		{
			t = haystack;
			t = strstr(t, "GET ");
			if(t)
				{
					t += 4;	// climb "GET "
					now = strstr(t, " HTTP");
					if(!now)
						return NULL;
					now +=2;
					strncpy(temp_get,t,((now-t))>GET_STR_SIZE ? GET_STR_SIZE: (now-t));
					get_list_add(temp_get,th,ip);
				}
		 }
	return NULL;
}

#define STRING_TEXT_PRIV(m) ((struct xt_string_info *) m)

static int checkentry(const char *tablename,
		      const void *ip,
		      const struct xt_match *match,
		      void *matchinfo,
		      unsigned int hook_mask)
{
	struct xt_string_info *conf = matchinfo;
	struct ts_config *ts_conf;

	/* Damn, can't handle this case properly with iptables... */
	if (conf->from_offset > conf->to_offset)
		return 0;
	if (conf->algo[XT_STRING_MAX_ALGO_NAME_SIZE - 1] != '\0')
		return 0;
	if (conf->patlen > XT_STRING_MAX_PATTERN_SIZE)
		return 0;
	ts_conf = textsearch_prepare(conf->algo, conf->pattern, conf->patlen,
				     GFP_KERNEL, TS_AUTOLOAD);
	if (IS_ERR(ts_conf))
		return 0;

	conf->config = ts_conf;

	return 1;
}

static void destroy(const struct xt_match *match, void *matchinfo)
{
	textsearch_destroy(STRING_TEXT_PRIV(matchinfo)->config);
}

static struct xt_match string_match= {
		.name 		= "string",
		.family		= AF_INET,
		.checkentry	= checkentry,
		.match 		= match,
		.destroy 	= destroy,
		.matchsize	= sizeof(struct xt_string_info),
		.me 		= THIS_MODULE
};

static int __init xt_string_init(void)
{
	return xt_register_match(&string_match);
}

static void __exit xt_string_fini(void)
{
	xt_unregister_match(&string_match);
}

module_init(xt_string_init);
module_exit(xt_string_fini);
